
var browserWindow = Components.classes["@mozilla.org/appshell/window-mediator;1"]
                   .getService(Components.interfaces.nsIWindowMediator)
                   .getMostRecentWindow("navigator:browser");
                   
const S = browserWindow.SoftwareTranslatorForWeb;

S.improveNative(String, Array);

function $(id)
{
    return document.getElementById(id);
}


const INF = Number.POSITIVE_INFINITY;

var $source, $target, $annotations, $sourceMsg, $targetMsg;


var dict = window.arguments[0];
var pdef = window.arguments[1];
var index = window.arguments[2];

var treeView = S.getCurrentTree()._STFW_view;

var def, filter = treeView.filter;

if (pdef) {
    def = new S.SafeDefinition(pdef.source, pdef.target);
    def.annotations = pdef.annotations;
}
else if (filter) {
    var t = S.SafeDefinition.escape2(filter);
    def = new S.SafeDefinition(t, t);
}
else
    def = new S.SafeDefinition;
    

function code (node) {
//     return node.value.replace(/\n/g, ' ');
    return node.value.strip();
}


function onLoad ()
{
    $source = $('source');          
    $target = $('target');           
    $annotations = $('annotations');             
    $sourceMsg = $('sourceMsg');     
    $targetMsg = $('targetMsg');
    
    $source.addEventListener('blur', onSourceBlur, true);
    
    $source.value = def.source;
    $target.value = def.target;
    $annotations.value = def.annotations || '';
    
    syncTarget();
    syncSource();
    
//     $source.addEventListener('keydown', onKeyDown, false);
//     $target.addEventListener('keydown', onKeyDown, false);
}

function onKeyDown (event)
{
    if (event.keyCode == 13) {
        event.preventDefault();
        event.stopPropagation();
        doOK();
        if (index === undefined)
            window.close();
    }
}

function onSourceChange ()
{
    def.source = code($source);
    syncSource();
    syncTarget();
//     onTargetChange();
}

function onTargetChange ()
{
    def.target = code($target);
//     syncSource();
    syncTarget();
}

    
function syncSource ()
{
    var source = def.source;
    var msg, type;
    if ((msg = def.sourceErrorMessage))
        type = 'error';
    else
    if (filter && !def.testSource(filter)) {
        type = 'warn';
        msg = 'not matches with filter text!';
    }
    else 
    if (dict.isDefined(source) && (index === undefined || dict._map[source] !== treeView._list[index])) {
        type = 'warn';
        msg = 'this source already exists (will replace)';
    }
    else {
        type = 'info';
        var priority = def.priority;
        msg = 'Priority: ' + (priority == INF ? '∞' : priority);
    }
    $sourceMsg.value = msg;
    $sourceMsg.className = type;
}

function syncTarget ()
{
    var msg = '', type = 'hidden';
    if (!def.sourceErrorMessage) {
        if ((msg = def.targetErrorMessage))
            type = 'error';
        else {
            var n = def.countNotUsed();
            if (n > 0) {
                type = 'warn';
                msg = n+' variable not used inside target';
            }
        }
    }
    $targetMsg.value = msg;
    $targetMsg.className = type;
}

function checkPriorities (action)
{
    if (!filter) return true;
    var n = treeView.chackForPriorityConflictOfFiltered(def.priority);
    if (!n) return true;
    if (dict.isDefined(code($source)) || (index !== undefined && treeView._list[index].priority >= def.priority))
        --n;
    if (!n) return true;
    return action(n == 1 ? 
        'There is 1 other definition with the same or highr priority that matches with the filter!':
        'There are '+n+' other definitions with the same or highr priority that match with the filter!');
}


function onSourceBlur ()
{
    checkPriorities(S.alert);
}


function doOK ()
{
    if (!checkPriorities(S.confirmSave)) return false;
    $source.removeEventListener('blur', onSourceBlur, true);
    if (index === undefined || dict.isDefined(code($source)))
        dict.setDefinition(code($source), code($target), $annotations.value)
    else
        dict.changeDefinition(index, code($source), code($target), $annotations.value)
    S.syncContent();
}


function doCancel ()
{
    $source.removeEventListener('blur', onSourceBlur, true);
}


