(function(){with(SoftwareTranslatorForWeb){

const VALID_TYPES = 'sdf'

const RE = /(.*?)(?:(?:%(?:(%)|(?:\((\w+)\))?(\w)|(.*)))|$)/g;
const RE_STR = 1;
const RE_ESCAPED = 2;
const RE_NAME = 3;
const RE_TYPE = 4;
const RE_ERROR = 5;

const INF = Number.POSITIVE_INFINITY;

const E = HASH();

newClass('DefinitionError', MyError);

newClass('Definition',
[
    function Definition (aSource, aTarget)
    {
        this._sourceMap = E;
        this._targetMap = E;
        this._priority = 0;
        this.source = aSource || '';
        this.target = aTarget || '';
    }
,    
    'getter',
    function exact () { return this._exact; }
,    
    'getter',
    function priority () { return this._priority; }
,    
    'getter',
    function source () { return this._source; }
,    
    'getter',
    function target () { return this._target; }
,    
    'setter',
    function source (value)
    {
        var tmap = HASH(), priority = 0, exact = true;
        var pat = '^', exactSource = '', names = [];
        var parts = this.parseDefinition(value);
        for (var i = 0, n = parts.length; i < n; ++i) {
            with (parts[i]) {
                if (str) {
                    pat += re.escape(str);
                    exactSource += str;
                }
                if (type)
                {
                    exact = false;
                    if (name in tmap)
                        throw new DefinitionError('variable "'+name+'" defined more then once');
                    names.push(name);
                    tmap[name] = type;
                    switch (type) {
                    case 's':
                        pat += '(.*)';
                        priority += 2;
                        break;
                    case 'd':
                        pat += '(\\d+)';
                        priority += 4;
                        break;
                    case 'f':
                        pat += '([\\d\\.\\,]+\\d+)';
                        priority += 3;
                        break;
                    default:
                        throw new Error('unexpected type "'+type+'"');
                    }
                }
            }
        }
        if (names.length > 9)
            throw new DefinitionError('too much variables: there are '+(names.length)+' variables, '+(9 - names.length)+' above the limit');
        if (exact) {
            this._priority = INF;
            this._exactSource = exactSource;
            this._re = null;
        } else {
            this._priority = priority;
            this._exactSource = null;
            this._re = new RegExp(pat+'$');
        }
        this._exact = exact;
        this._sourceMap = tmap;
        this._groupNames = names;
        return this._source = value;
    }
,    
    'setter',
    function target (value)
    {
        var tmap = HASH();
        var parts = this.parseDefinition(value);
        var repl = '', exactTarget = '';
        for (var i = 0, n = parts.length; i < n; ++i) {
            with (parts[i]) {
                if (str) {
                    repl += str.replace(/\$/g, '$$$$');
                    exactTarget += str;
                }
                if (type)
                {
                    if (name in tmap && tmap[name] != type)
                        throw new DefinitionError('variable "'+name+'" with different types');
                    tmap[name] = type;
                    repl += '$' + (this._groupNames.indexOf(name) + 1);
                }
            }
        }
        this._targetMap = tmap;
        this._repl = repl;
        this._exactTarget = exactTarget;
        return this._target = value;
    }
,    
    function testSource (string)
    {
        if (this._exact)
            return (string == this._exactSource);
        return this._re.test(string);
    }
, 
    function translate (string)
    {
        if (this._exact)
            return (string == this._exactSource) ? this._exactTarget : null;
        return string.replace(this._re, this._repl);
    }
,
    function parseDefinition (string)
    {
        var parts = [], m, r, index = 0;
        var regexp = RE;
        regexp.lastIndex = 0;
        while ( (m = regexp(string)) && m[0] )
        {
            if (m[RE_ERROR] !== undefined)
                throw new DefinitionError('invalid syntax');
                
            r = {};
                
            r.str = m[RE_STR] || '';
            if (m[RE_ESCAPED])
                r.str += m[RE_ESCAPED];
            
            if ((r.type = m[RE_TYPE])) {
                if (VALID_TYPES.indexOf(r.type) == -1)
                    throw new DefinitionError("invalid type '"+(r.type)+"', valid ones are: "+(VALID_TYPES.split('').join(', ')));
                r.name = m[RE_NAME] || (++index).toString();
            }
        
            parts.push(r);
        }
        return parts;
    }
,    
    function countNotUsed()
    {
        const sm = this._sourceMap;
        const tm = this._targetMap;
        var counter = 0;
        for (var name in sm)
            if (!(name in tm))
                ++counter;
        return counter;
    }
,    
    function findInconsistency()
    {
        const sm = this._sourceMap;
        const tm = this._targetMap;
        for (var name in tm) {
            if (!(name in sm))
                return 'variable "'+name+'" not defined into source';
            if (tm[name] != sm[name])
                return 'variable "'+name+'" with different types';
        }
        return null;
    }
,    
    'classmethod',
    function escape (string)
    {
        return string.strip().replace(/%/g, '%%');
    }
,    
    'classmethod',
    function escape2 (string)
    {
        return this.escape(string)
            .replace(/\d+[\.\,]+\d+/g, '%f')
            .replace(/\d+/g, '%d');
    }
]);



newClass('SafeDefinition', Definition,
[
    function SafeDefinition (aSource, aTarget, aAnnotation)
    {
        this._sourceErrorMessage = null;
        this._targetErrorMessage = null;
        this._inconsistenyErrorMessage = null;
        this._counter = 0;
        if (aAnnotation == 'undefined')
            aAnnotation = '';
        this.annotations = aAnnotation || '';
        parent.call(this, aSource, aTarget);
    }
,
    'getter',
    function isValid ()
    {
        return !this._sourceErrorMessage 
            && !this._targetErrorMessage
            && !this._inconsistenyErrorMessage;
    }
,
    'getter',
    function sourceErrorMessage ()
    {
        return this._sourceErrorMessage;
    }
,
    'getter',
    function targetErrorMessage ()
    {
        return this._targetErrorMessage || this._inconsistenyErrorMessage;
    }
,
    'getter',
    function counter ()
    {
        return this._counter;
    }
,
    'getter',
    function annotations ()
    {
        return this._annotations;
    }
,
    'setter',
    function annotations (value)
    {
        return (this._annotations = (/\S/.test(value) ? value : ''));
    }
,
    'setter',
    function source (value)
    {
        if (value != this._source)
            this._counter = 0;
        try
        {
            parent.call(this, value);
        }
        catch (e)
        {
            this._sourceErrorMessage = 'ERROR!';
            this._source = value;
            this._priority = 0;
            if (e instanceof DefinitionError) {
                this._sourceErrorMessage = e.message;
                return value;
            }
            throw e;
        }
        this._sourceErrorMessage = null;
        this._inconsistenyErrorMessage = this.findInconsistency();
        return value;
    }
,
    'setter',
    function target (value)
    {
        try
        {
            parent.call(this, value);
        }
        catch (e)
        {
            this._targetErrorMessage = 'ERROR!';
            this._target = value;
            this._priority = 0;
            if (e instanceof DefinitionError) {
                this._targetErrorMessage = e.message;
                return value;
            }
            throw e;
        }
        this._targetErrorMessage = null;
        this._inconsistenyErrorMessage = this.findInconsistency();
        return value;
    }
,
    function translate (string)
    {
        var rv = parent.call(this, string);
        this._counter++;
        return rv;
    }
]);

}})();
