(function(){with(SoftwareTranslatorForWeb){
const S = self;

newClass('DictionaryTagBuilder', TagBuilder)
.defineTags
([
    'definition','source','target','annotations','description'
]);

const INF = Number.POSITIVE_INFINITY;

const ASERV = Components.classes["@mozilla.org/atom-service;1"]
                .getService(Components.interfaces.nsIAtomService);
                
function atom (x) {
    return ASERV.getAtom(x);
//     return Components.classes["@mozilla.org/atom-service;1"]
//                 .getService(Components.interfaces.nsIAtomService)
//                 .getAtom(x);
}

newClass('TreeView',
{
    get rowCount () { return 0; }
}
,
[
    function getCellText (row, col) { return ''; },
    function setTree (treebox) { this._treeBox = treebox; },
    function isContainer () { return false; },
    function isSeparator () { return false; },
    function isSorted () { return false; },
    function getLevel (row) { return 0; },
    function getImageSrc (row, col) { return null; },
    function getRowProperties (row, props) {},
    function getCellProperties (row, col, props) {},
    function getColumnProperties (colid, col, props) {}
,
    function canDrop (index, orientation) { return false; },
    function cycleCell (row, col) {},
    function cycleHeader (col) {},
    function drop (row, orientation) {},
    function getCellValue (row, col) { return ''; },
    function getParentIndex (index) { return 0; },
    function performAction (action) {},
    function performActionOnCell (action, row, col) {},
    function performActionOnRow (action, row) {},
    function selectionChanged () {}    
]);


newClass('DictionaryTreeView', TreeView,
{
    get rowCount () { return this._list.length; }
,
    get filter () { return this._filter; }
,
    set filter (val)
    {
        val = val.strip();
        if (!val && !this._filter && this._list) return null;
        if (val === this._filter) return null;
        this._filter = val;
        this._sortedBy = 'priority';
        this._reversed = true;
        this._populate();
        return val;
    }
}
,
[
    function DictionaryTreeView (aDict, tree)
    {
        parent.call(this);
        
        this._sortedBy = null;
        this._reversed = null;
        this._list = [];
        this._filter = null;
        
        this._srcDict = aDict;
        this._treeBox = tree.boxObject;
        this._treeBox.QueryInterface(Components.interfaces.nsITreeBoxObject);
        
        aDict.registerView(this);
        this._populate();
        tree.view = this;
        tree._STFW_view = this;
        //this.filter = null;
    }          
,
    function getCellText (row, col)
    {
        var def = this._list[row];
        var val = def[col.id];
        if (col.id == 'priority' && val == INF)
            return '∞';
        return val;
    }
,
    function getCellProperties (row, col, props)
    {
        var def = this._list[row];
        switch (col.id) {
        case 'source':
            if (def.sourceErrorMessage)
                props.AppendElement(atom('withError'));
            break;
        case 'target':
            if (def.targetErrorMessage)
                props.AppendElement(atom('withError'));
            break;
        }
    }
,
    function cycleHeader (col, elem)
    {
        var dict = this._srcDict;
        var inv = (col.id == this._sortedBy) ? !this._reversed : false;
        this.sortBy(col.id, inv);
    }
,
    function invalidate (def)
    {
        var index = def ? this._list.indexOf(def) : null;
        if (index == -1) return;
        if (index !== null && !this._sortedBy)
            this._treeBox.invalidateRow(index)
        else
            this._populate();
            //this._treeBox.invalidateRange(0, this._list.length);
    }
,
    function _populate ()
    {
        var lst = [], map = this._srcDict._map;
        if (!this._filter) {
            for each (var def in map)
                lst.push(def);
        } else {
            var val = this._filter;
            for each (var def in map)
                if (!def.sourceErrorMessage && def.testSource(val))
                    lst.push(def);
        }
        this._treeBox.beginUpdateBatch();
        this._treeBox.rowCountChanged(0, -this._list.length);
        this._list = lst;
        this._treeBox.rowCountChanged(0, this._list.length);
        this._syncOrder();
        this._treeBox.endUpdateBatch();
    }
,
    function _defMatchFilter (def)
    {
        if (!this._filter) return true;
        return (!def.sourceErrorMessage && def.testSource(this._filter));
    }
,
    function sortBy (id, inv)
    {
        var sorted = this._sortedBy;
        this._sortedBy = id;
        this._reversed = inv;
        if (id || sorted)
            this._syncOrder();        
    }
,
    function dontSort ()
    {
        this.sortBy(null, false);
    }
,
    function _syncOrder ()
    {
        if (this._sortedBy) {
            var id = this._sortedBy;
            var cmp = (this._reversed) ?
                function(a, b) {
                    if (a[id] > b[id]) return -1;
                    if (a[id] < b[id]) return 1;
                    return 0;
                } :
                function(a, b) {
                    if (a[id] > b[id]) return 1;
                    if (a[id] < b[id]) return -1;
                    return 0;
                };
            this._list.sort(cmp);
            this._treeBox.invalidateRange(0, this._list.length);
            
            var dir = (this._reversed ? 'ascending' : 'descending');
            var cols = this._treeBox.element.firstChild.childNodes;
            for (var i = 0; i < cols.length; ++i)
                cols[i].setAttribute("sortDirection", (cols[i].id == id ? dir : 'natural'));
            
            return true;
        }
        else {
            var cols = this._treeBox.element.firstChild.childNodes;
            for (var i = 0; i < cols.length; ++i)
                cols[i].setAttribute("sortDirection", 'natural');
            
            return false;
        }
    }
,
    function chackForPriorityConflictOfFiltered (p)
    {
        var a = this._list.pluck('priority');
        a.sort();
        var i = a.length;
        p = p || a[i-1];
        var rv = 0;
        while (i-- && a[i] >= p) ++rv;
        return rv;
    }
]);


newClass('DictionaryError', MyError);

newClass('Dictionary',
{
    get revision ()
    {
        return this._revision;
    }
,
    get changed ()
    {
        return this._revision > this._era;
    }
,
    set changed (value)
    {
        if (value)
            this._revision++;
        else
            this._era = this.revision;
        return value;
    }
}
,
[
    function Dictionary ()
    {
        this._map = HASH();
        this._views = [];
        this._frized = 0;
        this._revision = 1
        this._era = this._revision;
    }
,
    function registerView (view)
    {
        this._views.push(view);
    }
,
    function savedToFile (file)
    {
        this.src = file.path;
        this.srcLastModifiedTime = file.exists ? file.lastModifiedTime : 0;
        this.changed = false;
    }
,
    function getBestDefinitionFor (s)
    {
        var def, map = this._map;
        var d = (s.indexOf('%') == -1) ? map[s] : null;
        if (d && d.exact && d.isValid) {
            def = d;
        } else {
            for each (d in map) {
                if (d.isValid && d.testSource(s) && (!def || d.priority > def.priority)) {
                    def = d;
                    if (def.priority == INF)
                        break;
                }
            }
        }
        return def;
    }
,
    function translate (string)
    {
        var rv;
        var m = string.match(STRIP_RE);
        string = m[2];
        var def = this.getBestDefinitionFor(string);
        if (!def) return null;
        try {
            rv = def.translate(string);
        } catch (err) {
            return null;
        }
        this.invalidateViews(def);
        return m[1] + rv + m[3];
    }
,
    function invalidateViews (def)
    {
        if (this._frized) return;
        this._views.invoke('invalidate', def);
    }
,
    function addDefinition (source, target, annotations)
    {
        if (!source) return;
        if (source in this._map)
            throw new DictionaryError('"'+source+'" source definition is already defined')
        var def = new SafeDefinition(source, target, annotations);
        this._map[source] = def;
        this._revision++;
        this.invalidateViews();
    }
,
    function setDefinition (source, target, annotations)
    {
        var def = this._map[source];
        if (!def) {
            this.addDefinition(source, target, annotations)
            return;
        }
        def.target = target;
        def.annotations = annotations;
        this._revision++;
        this.invalidateViews(def);
    }
,
    function changeDefinition (def, source, target, annotations)
    {
        var key = def.source;
        delete this._map[key];
        if (def.source != source)
            def.source = source;
        if (def.target != target)
            def.target = target;
        if (def.annotations != annotations)
            def.annotations = annotations;
        this._map[source] = def;
        this._revision++;
        this.invalidateViews();
    }
,
    function removeDefinitionBySource (source)
    {
        var def = this._map[source];
        if (def === undefined) return;
        delete this._map[source];
        this._revision++;
        this.invalidateViews();
    }
,
    function isDefined (source)
    {
        return (source in this._map);
    }
,
    function withNoTreeUpdates(fn, thisArg)
    {
        this._frized = true;
        try {
            return fn.call(thisArg);
        }
        finally {
            this._frized = false;
            this.invalidateViews();
        }
    }
,
    function toXML ()
    {
//         var doc = document.implementation.createDocument(dom.XHTMLNSURI, "dictionary", null);
//         var root = doc.documentElement;
//         with (new DictionaryTagBuilder(doc)) {
//             for each (var def in this._map) {
//                 root.appendChild(
//                     DEFINITION(
//                         SOURCE(def.source || ''),
//                         TARGET(def.target || ''),
//                         (def.annotations ? ANNOTATIONS(def.annotations) : null)
//                     )
//                 );
//             }
//         }
//         var serializer = new XMLSerializer();
//         return '<?xml version="1.0" encoding="UTF-8"?>' + serializer.serializeToString(doc);
        
        var xml = '<?xml version="1.0" encoding="UTF-8"?>\n';
        xml += '<dictionary xmlns="http://www.w3.org/1999/xhtml">\n';
        for each (var def in this._map) {
            xml += '<definition>\n';
            xml += '  <source>' + dom.escape(def.source || '') + '</source>\n';
            xml += '  <target>' + dom.escape(def.target || '') + '</target>\n';
            if (def.annotations)
                xml += '  <annotations>' + dom.escape(def.annotations || '') + '</annotations>\n';
            xml += '</definition>\n';
        }
        xml += '</dictionary>\n';
        return xml;
    }
,
    function toText ()
    {
        var txt = '';
        const map = this._map;
        for (var key in map) {
            var def = map[key];
            txt += buildTxtLine(def.source, def.target, def.annotations).strip() + '\r\n'
        }
        return txt;
    }
,
    function _createDefinition (source, target, annotations)
    {
        if (!source || source in this._map) return null;
        var def = new SafeDefinition(source, target, annotations);
        this._map[source] = def;
        return def;
    }
,
    'classmethod',
    function fromXML (xml)
    {
        var doc;
        if (typeof xml == 'string') {
            var parser = new DOMParser();
            var doc = parser.parseFromString(xml, "text/xml");
        } else {
            doc = xml;
        }
        var root = doc.documentElement;
        if (root.tagName == "parserError" || root.namespaceURI == "http://www.mozilla.org/newlayout/xml/parsererror.xml") {
            var msg = root.firstChild.textContent;
            if (msg.endswith(':'))
                msg = msg.substring(0, msg.length-1); 
            throw new DictionaryError(msg);
        }
        var dict = new this();
        var deftags = doc.getElementsByTagName('definition');
        for (var deftag, i = 0, n = deftags.length; i < n; ++i) {
            deftag = deftags[i];
            var source = deftag.getElementsByTagName('source')[0].firstChild.data;
            var target = deftag.getElementsByTagName('target')[0].firstChild.data;
            var anntag = deftag.getElementsByTagName('annotations')[0]
                      || deftag.getElementsByTagName('description')[0];
            var annotations = anntag ? anntag.firstChild.data : '';
            dict._createDefinition(source.strip(), target.strip(), annotations);
        }
        return dict;
    }
,
    'classmethod',
    function fromText (txt)
    {
        var lines = txt.replace(//g, '').splitlines();
        var dict = new this();
        for (var i = 0, n = lines.length; i < n; ++i) {
            if (!lines[i].strip())
                continue;
            var parts = parseTxtLine(lines[i]);
            if (parts.length < 2)
                continue;
            dict._createDefinition(parts[0].strip(), parts[1].strip(), parts[2]);           
        }
        return dict;
    }
,
    'classmethod',
    function from (data)
    {
        try {
            return this.fromXML(data);
        } catch (e) {
            return this.fromText(data);
        }
    }
]);


const STRIP_RE = /^(\s*)([\s\S]*?)(\s*)$/;

const FIRST_FIELD_FROM_TXT_LINE_RE = /^(?:"((?:""|[^"])*)"|([^\t]*))\t?(.*)/;

function parseTxtLine (line)
{
    var m, res = [];
    while (line) {
        m = FIRST_FIELD_FROM_TXT_LINE_RE(line);
        if (!m) break;
        if (m[1] !== undefined)
            res.push(m[1].replace(/""/g, '"'));
        else
            res.push(m[2]);
        line = m[3];
    }
    return res;
}

function buildTxtLine ()
{
    return Array.map(arguments, toTXTField).join('\t');
}

function toTXTField (s)
{
    if (/^\s|\s$|["&\t]/.test(s)) {
        s = '"' + s.replace(/\"/g, '""') + '"';
    }
    return s;
}

}})();
