(function(){
with (SoftwareTranslatorForWeb) {
with (newModule('dom')) {

const SP = ' ';

dom.XHTMLNSURI = "http://www.w3.org/1999/xhtml";

dom.FRAME_TAGS =
{
    "FRAME": 1,
    "IFRAME": 1
};

dom.FORCED_TRIM_TAGS =
{
    "LI": 1
}

dom.COMPLEX_TAGS =
{
    "APPLET": 1,
    "BR": 1,
    "EMBED": 1,
    "FRAME": 1,
    "FRAMESET": 1,
    "IFRAME": 1,
    "IMG": 1,
    "INPUT": 1,
//     "LABEL": 1,
    "NOEMBED": 1,
    "OBJECT": 1,
    "SELECT": 1,
    "TABLE": 1,
    "TEXTAREA": 1
};

dom.INVISIBLE_TAGS =
{
    "HTML": 1,
    "HEAD": 1,
//     "TITLE": 1,
    "META": 1,
    "LINK": 1,
    "STYLE": 1,
    "SCRIPT": 1,
//     "NOSCRIPT": 1,
    "BR": 1
};
    
addFunctions
([
    function $ (id, doc)
    {
        return (doc||document).getElementById(id);
    }
,
    function escape (s)
    {
        return s.replace(/&/g, "&amp;").replace(/</g, "&lt;").replace(/>/g, "&gt;").replace(/"/g, "&quot;").replace(/'/g, "&apos;");
    }
,
    function unescape (s)
    {
        return s.replace(/&apos;/g,"'").replace(/&quot;/g,"\"").replace(/&gt;/g,">").replace(/&lt;/g,"<").replace(/&amp;/g,"&");
    }
,
    function evaluate (xpath, root)
    {
        var doc = root ? (root.evaluate ? root : root.ownerDocument) : document;
        root = root || doc;
        var ss = doc.evaluate(xpath, root, null, 7, null);
        var n = ss.snapshotLength;
        var fn = arguments[2];
        if (!fn) {
            var result = new Array(n);
            while (n--) result[n] = ss.snapshotItem(n);
            return result;
        } else {
            var thisArg = arguments[3];
            for (var i = 0; i < n; ++i)
                fn.call(thisArg, ss.snapshotItem(i));
        }
        return ss;
    }
,
    function getSubWindows (node, deep)
    {
        var result;
        if (deep) {
            result = [];
            iterateWindows(node, result.push, result);
            result.removeAll(node);
        }
        else if (node instanceof Window) {
            result = Array.slice(node.frames);
            result.removeAll(node);
        }
        else {
            result = [];
            for (var tagName in FRAME_TAGS) {
                var frames = node.getElementsByTagName(tagName);
                for (var i = 0; i < frames.length; ++i)
                    result.push(frames[i].contentWindow);
            }
        }
        return result;
    }
,
    function getSubDocuments (node, deep)
    {
        return getSubWindows(node, deep).pluck('document');
    }
,
    function iterateWindows (node, handler, thisArg)
    {
        if (!node) return;
        
        if (node instanceof Window)
            handler.call(thisArg, node);
        
        var subWindows = getSubWindows(node);
        
        for (var i = 0; i < subWindows.length; ++i)
            iterateWindows(subWindows[i], handler, thisArg);
    }
,
    function setAttributes (node, attributes)
    {
        for (var name in attributes) {
            var value = attributes[name];
            if (value !== undefined)
                node.setAttribute(name, value);
        }
    }
,
    function append (node, child)
    {
        if (!child) return;
        if ( isArray(child) ) {
            for (var i = 0, n = child.length; i < n; ++i) {
                append(node, child[i]);
            }
        } else {
            if (typeof child == 'string')
                child = node.ownerDocument.createTextNode(child);
            node.appendChild(child);
        }
        return node;
    }
,
    function hasClass (node, name)
    {
        if (!node.className)
            return false;
        else
        {
            var re = new RegExp("(^|\\s)"+name+"($|\\s)");
            return re.test(node.className);
        }
    }
,
    function setClass (node, name)
    {
        if (node.className) {
            if (!hasClass(node, name))
                node.className += " " + name;
        }
        else
            node.className = name;
    }
,
    function removeClass (node, name)
    {
        if (!node.className) return;
        var re = new RegExp("(^|\\s+)"+name+"($|\\s)", 'g');
        node.className = node.className.replace(re, '$2');
    }
,
    function attrMap (node)
    {
        var map = {};
        var a = node.attributes, n = a.length;
        while (n--) map[a[n].name] = a[n].value;
        return map;
    }
,
    function firstChildWithName (node, name)
    {
        var child = node.firstChild;
        while (child && child.nodeName != name)
            child = child.nextSibling;
        return child;
    }
,
    function computedStyle (node, view)
    {
        return (view || node.ownerDocument.defaultView)
            .getComputedStyle(node, '');
    }
,
    function isInline (node, view)
    {
        return (view || node.ownerDocument.defaultView)
            .getComputedStyle(node, '').display == 'inline';
    }
, 
    function textNodeComputedValue (textNode)
    {
        if (!/\S/.test(textNode.data))
            return null;
        
        if (computedStyle(textNode.parentNode).whiteSpace == 'pre')
            return textNode.data;
            
        var text = textNode.data.replace(/\s+/g, SP);
            
//         if (text[0] == SP && _toTrimTextNode(textNode, prevInlineNode, /\s$/))
//             text = text.slice(1);
//         
//         if (text.slice(-1) == SP && _toTrimTextNode(textNode, nextInlineNode))
//             text = text.slice(0, -1);
        
        return text;
    }
,
    function prevNode (node)
    {
        var r = node.previousSibling;
        if (r === null)
            return node.parentNode;
        while (r.lastChild)
            r = r.lastChild;
        return r;
    }
,
    function nextNode (node)
    {
        var r = node.firstChild || node.nextSibling;
        while (r === null) {
            node = node.parentNode;
            if (node === null)
                return null;
            r = node.nextSibling;
        }
        return r;
    }
,
    function isTextNode (node)
    {
        try {
            return node && node.nodeType == 3;
        } catch (e) {
            return false;
        }
    }
,
    function prevInlineNode (node)
    {
        var r = node.previousSibling;
        if (r === null) {
            node = node.parentNode;
            if (node === null || !isInline(node))
                return null;
        } else {
            while (r) {
                if (r.nodeType == 1 && !isInline(r))
                    return prevInlineNode(r);
                node = r;
                r = r.lastChild;
            }
        }
        return node;
    }
,
    function nextInlineNode (node)
    {
        var r = node.firstChild || node.nextSibling;
        while (r === null) {
            node = node.parentNode;
            if (node === null || !isInline(node))
                return null;
            r = node.nextSibling;
        }
        if (r && r.nodeType == 1 && !isInline(r))
            return nextInlineNode(r);
        return r;
    }
]);

function _toTrimTextNode (node, getNext, test)
{
    while ((node = getNext(node))) {
        if (node.nodeType == 3) {
            if (/\S/.test(node.data)) {
                return test && test(node.data);
            }
        }
        else 
        if (node.nodeType == 1) {
            if (node.nodeName in FORCED_TRIM_TAGS)
                return true;
            else
            if (node.nodeName in COMPLEX_TAGS)
                return false;
        }
    }
    return true;
}



S.newClass('TagBuilder',
[
    function TagBuilder (aDocument)
    {
        this._document = aDocument || document;
    }
,
    'getter',
    function document () { return this._document; }
,
    function create (tagName)
    {
        var tag = this.createElement(tagName);
        var i = 1, n = arguments.length;
        if (i < n && arguments[i] && arguments[i].constructor === Object)
            setAttributes(tag, arguments[i++]);
        while (i < n) append(tag, arguments[i++]);
        return tag;
    }
,
    function createElement (name)
    {
        return this.document.createElement(name);
    }
,
    'classmethod',
    function defineTags (tagNames)
    {
        var proto = this.prototype;
        Array.forEach(tagNames, function(tagName) {
            proto[tagName.toUpperCase()] = function() {
                var tag = this.createElement(tagName);
                var i = 0, n = arguments.length;
                if (i < n && arguments[i] && arguments[i].constructor === Object)
                    setAttributes(tag, arguments[i++]);
                while (i < n) append(tag, arguments[i++]);
                return tag;
            }
        });
        return this;
    }
]);



S.newClass('XULTagBuilder', TagBuilder,
[
    function createElement (tagName)
    {
        const doc = this.document;
        if (doc instanceof XULDocument)
            return doc.createElement(tagName);
        else
            return doc.createElementNS(XULLNSURI, 'xul:'+tagName);
    }
])
.defineTags
([
    'tree','treecols','treecol','treechildren','treeitem','treerow','treecell',
    'listbox','listitem','box','hbox','vbox','label'
]);


S.newClass('HTMLTagBuilder', TagBuilder,
[
    function createElement (tagName)
    {
        const doc = this.document;
        if (doc instanceof HTMLDocument)
            return doc.createElement(tagName);
        else
            return doc.createElementNS(XHTMLNSURI, 'html:'+tagName);
    }
])
.defineTags
([
    'a','br','div','form','h1','h2','h3','hr','frame','iframe',
    'img','input','li','ol','p','pre','select','span','strong',
    'table','tbody','td','textarea','tfoot','th','thead','tr','tt','ul'
]);



}}})();
