with (SoftwareTranslatorForWeb) {

newClass('PrefBranch',
[
    function PrefBranch (aRoot)
    {
        this._branch = Components.classes["@mozilla.org/preferences-service;1"]
                                 .getService(Components.interfaces.nsIPrefService)
                                 .getBranch(aRoot);
        this._branch.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
        
        this.constructor = this.constructor;
        this._branch.__proto__ = this.constructor.prototype;
        this.__proto__ = this._branch;
    }
,
    function getSubBranch (aSubpath) {
        return new this.constructor(this.root + aSubpath);
    }
,
    function getUnicharPref (aPrefName) {
        return this.getComplexValue(aPrefName, 
            Components.interfaces.nsISupportsString).data;
    }
,
    function setUnicharPref (aPrefName, aValue) {
        var str = Components.classes["@mozilla.org/supports-string;1"]
                            .createInstance(Components.interfaces.nsISupportsString);
        str.data = aValue;
        this.setComplexValue(aPrefName, Components.interfaces.nsISupportsString, str);
    }
,
    function getLocalizedUnicharPref (aPrefName) {
        return this.getComplexValue(aPrefName, Components.interfaces.nsIPrefLocalizedString).data;
    }
,
    function setFilePref (aPrefName, aValue) {
        this.setComplexValue(aPrefName, Components.interfaces.nsILocalFile, aValue);
    }
,
    function getFilePref (aPrefName) {
        return this.getComplexValue(aPrefName, Components.interfaces.nsILocalFile);
    }
,
    function setRelFilePref (aPrefName, aValue, aRelTo) {
        var relFile = Components.classes["@mozilla.org/pref-relativefile;1"]
                                .createInstance(Components.interfaces.nsIRelativeFilePref);
        relFile.relativeToKey = aRelTo;
        relFile.file = aValue;
        this.setComplexValue(aPrefName, Components.interfaces.nsIRelativeFilePref, relFile);
    }
,
    function getRelFilePref (aPrefName) {
        return this.getComplexValue(aPrefName, Components.interfaces.nsIRelativeFilePref).file;
    }
,
    function clearUserPrefSafe (aPrefName) {
        try {
            this.clearUserPref(aPrefName);
        } catch(e) {}
    }
,
    function getJSPref (name) {
        return eval( '(' + this.getUnicharPref(name) + ')' );
    }
,
    function setJSPref (name, value) {
        var source = (typeof value == 'object')
                    ? value.toSource()
                    : [value].toSource().slice(1, -1);
        this.setUnicharPref(name, source);
    }
,
    function getCharListPref (name) {
        var t = this.getCharPref(name).strip();
        return (t ? t.split(/(\s*\,\s*)+/) : []);
    }
,
    function setCharListPref (name, value) {
        this.setCharPref(name, value.join(', '));
    }
,
    function getUnicharListPref (name) {
        var t = this.getUnicharPref(name).strip();
        return (t ? t.split(/(\s*\,\s*)+/) : []);
    }
,
    function setUnicharListPref (name, value) {
        this.setUnicharPref(name, value.join(', '));
    }
,
    function getWordsPref (name) {
        var t = this.getUnicharPref(name).strip();
        return (t ? t.split(/\W+/) : []);
    }
,
    function setWordsPref (name, value) {
        this.setUnicharPref(name, value.join(', '));
    }
,
    function getListPref (name) {
        var t = this.getUnicharPref(name);
        return (t ? t.split('|') : []);
    }
,
    function setListPref (name, value) {
        this.setUnicharPref(name, value.join('|'));
    }
,
    function getBoolListPref (name) {
        var t = this.getCharPref(name).strip();
        return (t ? t.split('').map(Number).map(Boolean) : []);
    }
,
    function setBoolListPref (name, value) {
        this.setCharPref(name, value.map(Number).join(''));
    }
]);

newClass('PrefObserver',
[
    function PrefObserver (branch, prefTypeDefs)
    {
        this._branch = branch;
        this._observing = false;
        this.__buildPrefTypeMaps(prefTypeDefs);
    }
,
    'getter',
    function branch () { return this._branch; }
,
    'getter',
    function observing () { return this._observing; }
,
    function register ()
    {
        if (!this._observing) {
            this._branch.addObserver("", this, false);
            this._observing = true;
        }
        return this;
    }
,
    function unregister ()
    {
        this._branch.removeObserver("", this);
        this._observing = false;
        return this;
    }
,
    function observe (aSubject, aTopic, aData)
    {
        if (aTopic != "nsPref:changed") return;
        this.onChange(aData, this.get(aData));
    }
,
    function onChange (name, value)
    {
    }
,
    function get (name)
    {
        return this._prefGetters[name].call(this._branch, name);
    }
,
    function set (name, value)
    {
        this._prefSetters[name].call(this._branch, name, value);
        return value;
    }
,
    function forEach (listOrPrefix, fn, thisArg)
    {
        var cl = listOrPrefix || '';
        if (typeof cl == 'string')
            cl = this._branch.getChildList(cl, {});
        for (var i = 0, n = cl.length; i < n; ++i) {
            fn.call(thisArg, cl[i], this.get(cl[i]));
        }
    }
,
    function refresh (arg)
    {
        this.forEach(arg, this.onChange, this);
    }      
,
    function typeOf (name)
    {
        return this._prefTypeDefs[name];
    }
,
    function __buildPrefTypeMaps (prefTypeDefs)
    {
        const td = this._prefTypeDefs = prefTypeDefs || {};
        this._prefGetters = {};
        this._prefSetters = {};
        const branch = this._branch;

        function defTypeOf (name) {
            switch (branch.getPrefType(name)) {
                case branch.PREF_STRING: return 'Char';
                case branch.PREF_INT: return 'Int';
                case branch.PREF_BOOL: return 'Bool';
            }
        }
        
        var childList = branch.getChildList('', {});
    
        for (var name, type, method, i = 0; i < childList.length; ++i){
            name = childList[i];
            td[name] = td[name] || defTypeOf(name);
            method = 'get'+td[name]+'Pref';
            if (!(method in branch)) throw new Error('method "'+method+'" not defined');
            this._prefGetters[name] = branch[method];
            method = 'set'+td[name]+'Pref';
            if (!(method in branch)) throw new Error('method "'+method+'" not defined');
            this._prefSetters[name] = branch[method];
        }
    }
]);

} // with
