(function(){with(SoftwareTranslatorForWeb){
const S = self;

newClass('BaseInspectorTool',
[
    function BaseInspectorTool (window)
    {
        this._window = window;
        var handlers = this._handlers = {};
        for (var key in this)
            if (/^on[A-Z]/.test(key))
                handlers[key.substring(2).toLowerCase()] = bind(this, this[key]);
    }
,
    'getter',
    function running () { return this._isRunning; }
,
    'setter',
    function running (val)
    {
        if (this._isRunning === !val) {
            if (val)
                this.start();
            else
                this.stop();
        }
        return val;
    }
,
    function start ()
    {
        this._registerHandlers();
        this._isRunning = true;
    }
,
    function stop ()
    {
        this._unregisterHandlers();
        this._isRunning = false;
    }
,
    function _registerHandlers ()
    {
        var handlers = this._handlers;
        dom.iterateWindows(this._window, function(target){
            for (var name in handlers)
                target.addEventListener(name, handlers[name], true);
        });
    }
,
    function _unregisterHandlers ()
    {
        var handlers = this._handlers;
        dom.iterateWindows(this._window, function(target){
            for (var name in handlers)
                target.removeEventListener(name, handlers[name], true);
        });
    }
,
    function onClick (event) { abortEvent(event); },
    function onMouseDown (event) { abortEvent(event); },
    function onMouseUp (event) { abortEvent(event); }
]);



newClass('SelectTextTool', BaseInspectorTool,
[
    function onKeyPress (event)
    {
        if (event.keyCode == event.DOM_VK_ESCAPE)
            S.enableSelectTextToTranslate(false);
    }            
,
    function _getNodeFromEvent (event, forced)
    {
        var node = event.explicitOriginalTarget;
        if (!node) return null;
        switch (node.nodeName) {
        case '#text':
        case 'INPUT':
        case 'IMG':
            return node;
        default:
            if (forced) {
                var res = dom.evaluate('.//text()', node);
                if (res.length == 1)
                    return res[0];
            }
            return null;       
        }
    }
,
    function onMouseMove (event)
    {
        var node = this._getNodeFromEvent(event, false);
        if (node) {
            try {
                this._selectNode(node);
            } catch (err) {
            }
        }
    }
,
    function onMouseOut (event)
    {
        var doc = event.originalTarget.ownerDocument;
        var win = doc.defaultView;
        var s = win.getSelection();
        if (s.rangeCount > 0) s.removeAllRanges();
    }
,
    function _selectNode (node)
    {
        var doc = node.ownerDocument;
        var win = doc.defaultView;
        var s = win.getSelection();
        if (s.rangeCount > 0) s.removeAllRanges();
        var range = doc.createRange();
        range.selectNode(node);
        s.addRange(range);
    }
,
    function onMouseDown (event)
    {
        var node = this._getNodeFromEvent(event, true);
        if (!node) return;
        abortEvent(event);
        var value = node._SoftwareTranslatorForWeb_original
                 || S.getNodeValue(node);
        if (!value) return;
        var filter;
        if (node.nodeType == 1) {
            if (event.ctrlKey || event.metaKey)
                this.openPopupSelector(event, value);
            else
                filter = value.value || value.src;
        } else {
            filter = value;
        }
        if (filter)
            S.setFilter(filter, true);
    }
,
    function openPopupSelector (event, data)
    {
        var popup = dom.$('SoftwareTranslatorForWeb-popup');
        while (popup.firstChild)
            popup.removeChild(popup.firstChild);
        var keys = [];
        for (var key in data)
            keys.push(key);
        keys.sort();
        keys.forEach(function(key) {
            var value = data[key];
            var item = document.createElement("menuitem");
            item.setAttribute('style', "font-weight:bold;");
            item.setAttribute('label', key);
            item.setAttribute('tooltiptext', data[key]);
            item.addEventListener('command', function(event) {
                S.setFilter(value, true);
            }, false);
            popup.appendChild(item);
        });
        openPopupAtScreen(popup, event.screenX, event.screenY, true);
    }
]);


}})()
